/*
A demonstration of a Gemini-native UI for long-form text entry.
Copyright (C) 2021 Lykso

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef DEBUG
#define DEBUG false
#endif

#include <dirent.h>
#include <errno.h>
#include <libgen.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "./config.h"
#include "./lib/common.h"
#include "./lib/ser_ll.h"
#include "./vendor/URLDecode/urldecode.h"

int show_homepage() {
  char *hostname = getenv("SERVER_NAME");

  printf(
      "20 text/gemini\r\n"
      "Welcome to %s. This is a UI experiment. As such, please don't use it for "
      "anything serious or get too attached to it. It may break and it may all "
      "get erased in the course of things.\r\n"
      "If you have feedback, email me at gemini@lyk.so.\r\n"
      "=> /?edit Set a client certificate and click here to get started.\r\n"
      "## Current features:\r\n"
      "* Long-form entry editing.\r\n"
      "## Other UI experiments:\r\n"
      "=> gemini://alpha.lyk.so\r\n"
      "=> gemini://beta.lyk.so\r\n"
      , hostname
      );

  return 0;
}

bool toggles_preformatted_mode(char* chunk) {
  char* substr = chunk;
  bool toggles = begins_with(substr, "```");

  while ((substr = strstr(substr, "\n```"))) {
    substr++;
    toggles = !toggles;
  }
  return toggles;
}

int show_page(bool editable, const char* path)
{
  DIR *dir;
  struct dirent *ent;

  if (is_list(path)) {
    ListMeta meta = {};

    if (!read_meta(&meta, path)) {
      printf("40 Unable to load page metadata!\r\n");
      return 0;
    }
    printf("20 text/gemini\r\n");

    if (DEBUG && editable) {
        printf("# Meta\r\nHead ID: %d\r\nTail ID: %d\r\nFree ID:%d\r\nNext ID: %d\r\nList Dir: %s\r\n--\r\n",
            meta.head_id, meta.tail_id, meta.free_id, meta.next_id, meta.list_dir);
    }

    ListNode node = {};
    NODE_ID_TYPE next_node_id = meta.head_id;
    bool in_preformatted_mode = false;

    while (next_node_id) {
      if (!read_node(&meta, &node, next_node_id)) {
        printf(
            "Oops! Failed to load chunk #%d! "
            "Please report this error to gemini@lyk.so.\r\n",
            next_node_id);
        return 0;
      }

      // Make sure the string terminates before using it.
      node.content[INPUT_LEN - 1] = '\0';

      char* decoded_content = urlDecode(node.content);

      if (editable) {
        if (DEBUG) {
          printf("Prev ID: %d\r\nNode ID: %d\r\nNext ID: %d\r\n",
              node.prev_id, node.id, node.next_id);
        }
        /*
        */

        // Continue preformatted mode from previous chunk.
        if (in_preformatted_mode) {
          printf("```\r\n");
        }

        if (toggles_preformatted_mode(decoded_content)) {
          in_preformatted_mode = !in_preformatted_mode;
        }

        printf("%s\r\n", decoded_content);

        // Leave preformatted mode if this chunk left us in it.
        if (in_preformatted_mode) {
          printf("```\r\n");
        }

        printf("=> /after/%d Insert after chunk.\r\n", node.id);
        printf("=> /before/%d Insert before chunk.\r\n", node.id);

        if (meta.head_id != node.id) {
          printf("=> /up/%d Move chunk up.\r\n", node.id);
        }

        if (meta.tail_id != node.id) {
          printf("=> /down/%d Move chunk down.\r\n", node.id);
        }
        printf("=> /replace/%d Replace chunk.\r\n\r\n", node.id);
        printf("=> /delete/%d DELETE chunk.\r\n", node.id);
        printf("--\r\n");

      } else {
        printf("%s", urlDecode(node.content));
      }
      next_node_id = node.next_id;
    }
    printf("\r\n");

    if (editable) {
      printf("=> /append Append chunk.\r\n");
      printf("=> /view View page.\r\n");
    }

  } else {
    printf("20 text/gemini\r\n");

    if (editable) {
      printf("=> /append Click to begin writing.\r\n");

    } else {
      printf("20 text/gemini\r\n");
      printf("No luck, sorry friend.\r\n");
    }
  }

  return 0;
}

int do_view(char* user)
{
  size_t data_path_len = sizeof(PAGES_DIR) + strlen(user) + 2;
  char* data_path = calloc(data_path_len, sizeof(char));
  snprintf(data_path, data_path_len, "%s/%s", PAGES_DIR, user);

  show_page(false, data_path);

  free(data_path);
  return 0;
}

int do_authed_view(char* data_path, const char* user) {
  show_page(false, data_path);
  printf("--\r\n=> /?edit Return to editing\r\n");

  char* remainder = strchr(user, ':');

  if  (!remainder) {
    printf("=> /view/%s Shareable URL.\r\n", user);

  } else {
    char* shareable = calloc(strlen(user) + 4, sizeof(char));
    memcpy(shareable, user, remainder - user);
    memcpy(shareable + (remainder - user), "%3A", 3);
    memcpy(shareable + (remainder - user) + 3, remainder+1, strlen(remainder+1));
    printf("=> /view/%s Shareable URL.\r\n", shareable);
    free(shareable);
  }
  return 0;
}

int do_edit(char* data_path, const char* user)
{
  show_page(true, data_path);

  if (user) {
    printf("=> /view/%s Shareable URL.\r\n", user);
  }
  return 0;
}

int do_append(char* data_path, char* chunk)
{
  if (strempty(chunk)) {
    printf("10 Enter text to append:\r\n");
    return 0;
  }

  if (!ensure_list(data_path)) {
    printf("40 Something went wrong while trying to get things set up...\r\n");
    return 0;
  }

  ListMeta meta = {};
  if (!read_meta(&meta, data_path)) {
    printf("40 Failed trying to open page meta file!\r\n");
    return 0;
  }

  ListNode node = {};
  init_node(&meta, &node);
  snprintf(node.content, INPUT_LEN, "%s", chunk);

  if (!append_node(&meta, &node)) {
    printf("40 Something went wrong while appending!\r\n");
    return 0;
  }

  //printf("20 text/plain\r\n%s\r\n%s\r\n", chunk, urlDecode(chunk));
  printf("30 /?edit\r\n");
  return 0;
}

int do_delete(ListMeta* meta, ListNode* node)
{
  if (!remove_node(meta, node)) {
    printf("40 Something went wrong while deleting!\r\n");
    return 0;
  }

  printf("30 /?edit\r\n");
  return 0;
}

int do_down(ListMeta* meta, ListNode* node)
{
  if (node->next_id == 0) {
    printf("40 Chunk is already at the bottom!\r\n");
    return 0;
  }

  // Since remove_node mutates the state of its arguments...
  NODE_ID_TYPE next_node_id = node->next_id;

  if (!remove_node(meta, node)) {
    printf("40 Something went wrong while moving the chunk!\r\n");
    return 0;
  }

  ListNode next_node = {};
  if (!read_node(meta, &next_node, next_node_id)) {
    printf("40 Error reading neighbor chunk!\r\n");
    return 0;
  }

  if (!insert_after(meta, &next_node, node)) {
    printf("40 Something went wrong while moving the chunk!\r\n");
    return 0;
  }

  printf("30 /?edit\r\n");
  return 0;
}

int do_up(ListMeta* meta, ListNode* node)
{
  if (node->prev_id == 0) {
    printf("40 Chunk is already at the top!\r\n");
    return 0;
  }

  // Since remove_node mutates the state of its arguments...
  NODE_ID_TYPE prev_node_id = node->prev_id;

  if (!remove_node(meta, node)) {
    printf("40 Something went wrong while moving the chunk!\r\n");
    return 0;
  }

  ListNode prev_node = {};
  if (!read_node(meta, &prev_node, prev_node_id)) {
    printf("40 Error reading neighbor chunk!\r\n");
    return 0;
  }

  if (!insert_before(meta, &prev_node, node)) {
    printf("40 Something went wrong while moving the chunk!\r\n");
    return 0;
  }

  printf("30 /?edit\r\n");
  return 0;
}

int do_before(ListMeta* meta, ListNode* node, char* chunk)
{
  if (strempty(chunk)) {
    printf("10 Enter text to insert:\r\n");
    return 0;
  }

  ListNode new_node = {};
  init_node(meta, &new_node);
  snprintf(new_node.content, INPUT_LEN, "%s", chunk);

  if (!insert_before(meta, node, &new_node)) {
    printf("40 Something went wrong while inserting!\r\n");
    return 0;
  }

  printf("30 /?edit\r\n");
  return 0;
}

int do_after(ListMeta* meta, ListNode* node, char* chunk)
{
  if (strempty(chunk)) {
    printf("10 Enter text to insert:\r\n");
    return 0;
  }

  ListNode new_node = {};
  init_node(meta, &new_node);
  snprintf(new_node.content, INPUT_LEN, "%s", chunk);

  if (!insert_after(meta, node, &new_node)) {
    printf("40 Something went wrong while inserting!\r\n");
    return 0;
  }

  printf("30 /?edit\r\n");
  return 0;
}

int do_replace(ListMeta* meta, ListNode* node, char* chunk)
{
  if (strempty(chunk)) {
    printf("10 Enter text to insert:\r\n");
    return 0;
  }

  snprintf(node->content, INPUT_LEN, "%s", chunk);

  if (!write_node(meta, node)) {
    printf("40 Something went wrong while replacing!\r\n");
    return 0;
  }

  printf("30 /?edit\r\n");
  return 0;
}


int main()
{
  char *query = getenv("QUERY_STRING");
  char *path_info = getenv("PATH_INFO");
  char *client_hash = getenv("TLS_CLIENT_HASH");

  // No null, please.
  if (!path_info) {
    path_info = "";
  }

  if (!query) {
    query = "";
  }

  // Drop the colon prefix on the client hash. Different algorithms should have
  // different lengths anyway, so the prefix is redundant and the colon causes
  // problems when linking.
  if (client_hash) {
    char* remainder = strchr(client_hash, ':');
    if (remainder) {
      client_hash = remainder+1;
    }
  }

  // "/"
  if (strempty(path_info) && !streq(query, "edit")) {
    return show_homepage();
  }

  // Redirect away from paths that end in '/' to simplify our code.
  if (path_info[strlen(path_info) - 1] == '/') {
      path_info[strlen(path_info) - 1] = '\0';
      printf("30 %s\r\n", path_info);
      return 0;
  }

  // Drop leading '/'.
  // gmid always includes this, but let's check in case that changes someday.
  if (*path_info == '/') {
    path_info++;
  }

  if (begins_with(path_info, "view/")) {
    return do_view(path_info + sizeof("view/") - 1);
  }

  // Authenticated routes follow.
  // Force authentication.
  if (strempty(client_hash)) {
    printf("60 \r\n");
    return 0;
  }

  // Calculate user's data directory path.
  size_t data_path_len = sizeof(PAGES_DIR) + strlen(client_hash) + 2;
  char* data_path = calloc(data_path_len, sizeof(char));
  snprintf(data_path, data_path_len, "%s/%s", PAGES_DIR, client_hash);

  if (streq(path_info, "append")) {
    return do_append(data_path, query);
  }

  if (strempty(path_info) && streq(query, "edit")) {
    return do_edit(data_path, client_hash);
  }

  // The following routes assume the user's linked list has been initiated.
  if (!is_list(data_path)) {
    printf("30 /?edit \r\n");
    return 0;
  }

  if (streq(path_info, "view")) {
    return do_authed_view(data_path, client_hash);
  }

  // The following routes operate on list nodes.
  char* node_id_str = strchr(path_info, '/')+1;

  if (strempty(node_id_str) || !is_all_digits(node_id_str) || atoi(node_id_str) == 0) {
    printf("50 Non-zero, numeric node ID required.\r\n");
    return 0;
  }

  NODE_ID_TYPE node_id = atoi(node_id_str);

  ListMeta meta = {};
  read_meta(&meta, data_path);

  ListNode node = {};
  read_node(&meta, &node, node_id);

  if (begins_with(path_info, "after/")) {
    return do_after(&meta, &node, query);
  }

  if (begins_with(path_info, "before/")) {
    return do_before(&meta, &node, query);
  }

  if (begins_with(path_info, "delete/")) {
    return do_delete(&meta, &node);
  }

  if (begins_with(path_info, "down/")) {
    return do_down(&meta, &node);
  }

  if (begins_with(path_info, "up/")) {
    return do_up(&meta, &node);
  }

  if (begins_with(path_info, "replace/")) {
    return do_replace(&meta, &node, query);
  }
  printf("40 Route not found.\r\n");


  return 0;
}